#! /usr/bin/env python3
"""
A module for analysing the data contained in a :class:`Segment` object.

Contains the :class:`Segment` class and methods for calculating various definitions of the effective mass.
"""

import warnings
import numpy as np
from scipy import integrate
from effmass import extrema
from effmass import angstrom_to_bohr
from effmass import ev_to_hartree
from effmass import boltzmann
from effmass import q

warnings.simplefilter(action='ignore', category=FutureWarning)


def _check_poly_order(polyfit_order):
    """Raises an AssertionError if the order of the polynomial is less than 2.

    Args:
        polyfit_order (int): the order of the polynomial fit.

    Returns:
        None.
    """

    if polyfit_order < 2:
        raise ValueError("the order of the polynomial must be 2 or higher")


def _solve_quadratic(a, b, c):
    r"""
    Solves quadratic equation of the form :math:`ax^2+bx+c=0` for multiple values of c.
    If the determinant is more than 0 (two solutions), it always returns the larger root.

    Args:
        a (float): the coefficient of :math:`x^2`.
        b (float): the coefficient of :math:`x^1`.
        c (list(float)): a list of coefficients of :math:`x^0`.

    Returns:
        list(float): the root of the equation for each value of c.
    
    """
    det = [b**2 - 4 * a * item for item in c]
    x = []
    for d in det:
        if d < 0:
            x.append(np.nan)
        if d == 0:
            x.append(-b / (2 * a))
        if d > 0:
            x.append((-b + np.sqrt(d)) / (2 * a))
    return x


class Segment:
    """Class for segments of the bandstructure. A Segment contains data for a
    particular region of reciprocal space and particular band.

    Attributes:
        band (int): The band number of the segment (counting starts from 0).
        kpoint_indices (list(int)): The indices of :attr:`effmass.inputs.Data.kpoints` from which this Segment is formed.
        kpoints (array(float)): 2-dimensional array. Each row contains the fractional coordinates of a kpoint [kx,ky,kz]. A slice of :attr:`effmass.inputs.Data.kpoints`.
        cartesian_kpoints (array(float)): 2-dimensional array. Each row contains the cartesian coordinates (angstrom :math:`^{-1}`) of a kpoint.
        dk_angs (array(float)): 1-dimensional array which contains the distance (angstrom :math:`^{-1}`) between each kpoint and the extrema.
        dk_bohr (array(float)): 1-dimensional array which contains the distance (bohr :math:`^{-1}`) between each kpoint and the extrema.
        energies (array(float)): 1-dimensional array which contains the energy (eV) of each eigenstate. A slice of :attr:`effmass.inputs.Data.energies`.
        dE_eV (array(float)): 1-dimensional array which contains the difference in energy (hartree) between each kpoint and the extrema.
        dE_hartree (array(float)): 1-dimensional array which contains the difference in energy (eV) between each kpoint and the extrema.
        occupancy (array(float)): 2-dimensional array. Each row contains occupation number of the eigenstates for a particular band. A slice of :attr:`effmass.inputs.Data.occupancy`.
        direction (array): 1-dimensional array with length 3. The direction between kpoints in the segment.
        band_type (str): The band type, determined by occupancy of the eigenstate. Argument choices are "conduction_band", "valence_band" or "unknown". If set to "unknown", some class methods will raise a warning and return None. 
        fermi_energy (float): the fermi energy in eV.
        dos (array): 2-dimensional array. Each row contains density of states data (units "number of states / unit cell")  at a given energy: [energy(float),dos(float)]. A slice of :attr:`effmass.inputs.Data.dos`.
        integrated_dos (array): 2-dimensional array. Each row contains integrated density of states data at a given energy: [energy(float),integrated_dos(float)]. A slice of :attr:`effmass.inputs.Data.integrated_dos`.
           

    """

    def __init__(self, Data, band, kpoint_indices):
        """Initialise an instance of the Segment class.

        Args:
            Data (Data): Data instance initialised from the bandstructure which contains the segment
            band (int): the band number of the segment
            kpoint_indices (list(int)): the kpoint indices of the segment

        Returns:
            None.
        """

        self.band = band
        self.kpoint_indices = kpoint_indices
        self.kpoints = np.array([Data.kpoints[k] for k in kpoint_indices
                                 ])  # in units 2*pi/angstrom
        self.cartesian_kpoints = np.array([
            np.dot(k, Data.reciprocal_lattice) for k in self.kpoints
        ])  # in units 1/Angstrom. Reciprocal lattice includes factor 2*pi.
        self.dk_angs = np.linalg.norm(
            self.cartesian_kpoints - self.cartesian_kpoints[0], axis=1)
        self.dk_bohr = np.divide(
            self.dk_angs, angstrom_to_bohr
        )  # divide as we are in reciprocal space --> units in inverse length
        self.energies = np.array(
            [Data.energies[band, k] for k in kpoint_indices])  # in units eV
        self.dE_eV = self.energies - self.energies[0]
        self.dE_hartree = np.multiply(self.energies - self.energies[0],
                                      ev_to_hartree)
        self.occupancy = np.array(
            [Data.occupancy[band, k] for k in kpoint_indices])
        self.direction = extrema.calculate_direction(self.kpoints[1],
                                                     self.kpoints[2])
        self.fermi_energy = Data.fermi_energy
        self.dos = self._dos(Data)
        self.integrated_dos = self._integrated_dos(Data)
        self._VBM = Data.VBM
        self._CBM = Data.CBM
        self.band_type = self._band_type()

    def __str__(self):
        """
        Segment string method.

        Returns:
            A string containing the energy of the Segment extrema (referenced to the VBM) and the start- and end- points of the Segment in reciprocal space.
        """
        energy_str = "{0:.2f}".format(self.energies[0] - self._VBM)
        start_str = str(np.round(self.kpoints[0],3))
        end_str = str(np.round(self.kpoints[-1],3))
        return energy_str+" eV; "+start_str+"-->"+end_str

    def _check_kanefit_points(self, polyfit_order=6):
        """Raises an AssertionError if there are not enough data points to fit
        a Kane dispersion.

        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.

        Returns:
            None.
        """
        idx = self.explosion_index(polyfit_order=polyfit_order)
        assert idx > 3,"Unable to calculate alpha parameter as inflexion too close to extrema"

    def explosion_index(self, polyfit_order=6):
        r"""
        This will find the index at which there is a change in sign of the second derivative. 

        In the region of this point the first derivative will pass through zero and so the transport mass (:math:`\frac{1}{\delta E \delta k}`) will explode.

        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.

        Notes:
            This marks the point at which the Kane dispersion is definitely not valid, although it may be that the Kane dispersion is a poor approximation prior to this.
        """
        dedk, d2edk2 = self.poly_derivatives(
            polyfit_order=polyfit_order,
            dk=self.dk_bohr,
            polyfit_weighting=False)
        sign = np.sign(d2edk2)
        signchange = ((np.roll(sign, 1) - sign) != 0).astype(int)
        signchange[0] = 0
        if 1 in signchange:
            cutoff = np.where(signchange == 1)[0][0]
        else:
            cutoff = len(self.dE_eV) - 1
        return cutoff

    def _fermi_function(self, eigenvalue, fermi_level=None, temp=300):
        r""" 
        Calculates the probability that an eigenstate is occupied using Fermi-Dirac statistics:
            ..math::
                p=\frac{1}{e^{\frac{Delta E}{kT}+1}

        Args:
            eigenvalue (float): the energy (eV) of the eigenstate.
            fermi_level (float): the fermi level (eV).

        Returns:
            float: the probability that the eigenstate is occupied
        
        """
        if temp <= 0:
            raise ValueError("temperature must be more than 0K")
        fermi_level = self.fermi_energy if fermi_level is None else fermi_level
        if self.band_type == "conduction_band":
            probability = (1 / ((np.exp(
                (eigenvalue - fermi_level) / (((boltzmann * temp) / q))) + 1)))
        elif self.band_type == "valence_band":
            probability = np.subtract(1,(1 / ((np.exp(
                (eigenvalue - fermi_level) / (((boltzmann * temp) / q))) + 1))))
        else:
            raise ValueError("Unable to calculate fermi function as the band type is unknown. Please set the Segment.band_type attribute manually (options are \"valence_band\" or \"conduction_band\").")
        return probability

    def weighting(self, fermi_level=None, temp=300):
        """Calculates a weighting for each kpoint using the Fermi-Dirac
        statistics.

        Args:
            quasi_fermi_level (float, optional): The fermi level to be used in Fermi-Dirac statistics. Defaults to :attr:`effmass.inputs.Segment.fermi_energy`.
            temp (float, optional): The temperature (K) to be used in Fermi-Dirac statistics. Defaults to 300.

        Returns:
            array(float): A 1-dimensional array which contains the weighting for each k-point.

        Notes:
            The weighting is relative only: constants will cancel out when using to weight least square fits or means.
        """

        fermi_level = self.fermi_energy if fermi_level is None else fermi_level
        # weighting is given by the fermi function
        weighting = (np.array([
            self._fermi_function(E, fermi_level=fermi_level, temp=temp)
            for E in self.energies
        ]))

        assert weighting.all() != 0,"Unable to assign a weighting to the dispersion as the Fermi-Dirac distribution at all kpoints is smaller than Python's double precision floats. Calculate band edge values using Segment.five_point_leastsq_effmass() or Segment.finite_difference_effmass()."

        return weighting

    def _band_type(self):
        """Identifies the band type, determined by the occupancy of the
        first k-point in the segment.

        An occupancy of 1.0 or 2.0 returns "valence_band" (as :class:`~effmass.analysis.Segment` is in the valence band).
        An occupancy of 0.0 returns "conduction_band" (as :class:`~effmass.analysis.Segment` is in the conduction band).
        In the case of partial occupancy, "unknown" is returned.

        Args:
            None

        Returns:
            str: the band type of the segment, either "valence_band", "conduction_band" or "unknown".
        """

        if self.energies[0] <= self._VBM:
            band_type = "valence_band"
        elif self.energies[0] >= self._CBM:
            band_type = "conduction_band"
        else:
            print(
                "The band type unknown. Please set the Segment.band_type attribute manually."
            )
            band_type = "unknown"
        return band_type

    def _dos(self, Data):
        """Returns slice of :attr:`effmass.Data.dos` corresponding to the
        energy range of the segment.

        Args:
            Data (Data): :class:`~effmass.inputs.Data` instance which was used to generate the :class:`~effmass.analysis.Segment`.

        Returns:
            list(floats): slice of :attr:`effmass.Data.dos` corresponding to the energy range of the segment.
        """
        if Data.dos == []:
            dos = []
        else:
            dos = []
            for i in range(len(self.energies)):
                for j in range(len(Data.dos)):
                    if self.energies[i] < Data.dos[j][0]:
                        dos.append(Data.dos[j][1])
                        break

        return dos

    def _integrated_dos(self, Data):
        """Returns slice of :attr:`effmass.Data.integrated_dos` corresponding
        to the energy range of the segment.

        Args:
            Data (Data): :class:`~effmass.inputs.Data` instance which was used to generate the :class:`~effmass.analysis.Segment`.

        Returns:
            array: slice of :attr:`effmass.Data.integrated_dos` corresponding to the energy range of the segment.
        """
        if Data.integrated_dos == []:
            integrated_dos = []
        else:
            integrated_dos = []
            for i in range(len(self.energies)):
                for j in range(len(Data.integrated_dos)):
                    if self.energies[i] < Data.integrated_dos[j][0]:
                        integrated_dos.append(Data.integrated_dos[j][1])
                        break
        return integrated_dos

    # The collection of methods below calculate the optical effective mass by integrating numerically the analytical
    # expression for the second derivative of the Kane dispersion multiplied by a Fermi-Dirac weighting.

    def mass_integration(self,
                         fermi_level=None,
                         temp=300,
                         alpha=None,
                         mass_bandedge=None,
                         upper_limit=None):
        """Integrates the product of the fermi-dirac distribution, density of
        states and second derivative of kane dispersion along the one-
        dimensional slice of k-space defined by
        :class:`~effmass.analysis.Segment` (up to
        :meth:`~effmass.analysis.Segment.explosion_index`).

        Args:
            fermi_level (float, optional): Fermi level (eV) to be used in Fermi-dirac statistics. Defaults to :attr:`~effmass.analysis.Segment.fermi_energy`.
            temp (float, optional): The temperature (K) to be used in Fermi-Dirac statistics. Defaults to 300.
            alpha (float, optional): The alpha parameter of the Kane dispersion (hartree$^{-1}$)
            mass_bandedge: The mass at bandedge parameter of the Kane dispersion (units electron mass).
            upper_limit (float, optional): The integration upper limit (bohr$^{-1}$). Defaults to where the Kane quasi-linear dispersion is no longer valide, defined by :meth:`~effmass.analysis.Segment.explosion_index`.

        Returns:
            float: The optical effective mass (units of electron mass), defined as the inverse of the second derivative of a kane dispersion, weighted according to occupancy of available eigenstates (the product of density of states and the fermi-dirac distribution).
        Note:
            The sign of the alpha parameter and mass_bandedge are important. If these are negative (as would be expected for the valence band), then they must be passed as negative values to the function.
        """

        alpha = self.alpha() if alpha is None else alpha
        mass_bandedge = self.kane_mass_band_edge(
        ) if mass_bandedge is None else mass_bandedge
        upper_limit = self.dk_bohr[
            self.explosion_index()] if upper_limit is None else upper_limit
        fermi_level = self.fermi_energy if fermi_level is None else fermi_level

        result = integrate.quad(
            self._mass_integrand,
            0,
            upper_limit,
            args=(fermi_level, temp, alpha, mass_bandedge))
        return result

    def _mass_integrand(self, k, fermi_level, temp, alpha, mass_bandedge):
        """Helper function for
        :meth:`~effmass.analysis.Segment.mass_integration`.

        A function for the weighted mass integrand.
        """
        return self.fd(k, fermi_level, temp, alpha, mass_bandedge) * (
            self._second_derivative_kane_dispersion(k, alpha, mass_bandedge))

    def fd(self, k, fermi_level, temp, alpha, mass_bandedge):
        r"""
        Calculates the probability that an eigenstate of momentum k is occupied, using Fermi-Dirac statistics and assuming a Kane dispersion.

        Args:
            k: the momentum of the eigenstate (bohr:math:`^{-1}`).
            fermi_level (float, optional): Fermi level (eV) to be used in Fermi-dirac statistics. Defaults to :attr:`~effmass.analysis.Segment.fermi_energy`.
            temp (float, optional): The temperature (K) to be used in Fermi-Dirac statistics. Defaults to 300.
            alpha (float, optional): The alpha parameter of the Kane dispersion (hartree$^{-1}$)
            mass_bandedge: The mass at bandedge parameter of the Kane dispersion (units electron mass).

        Returns:
            float: The probability that the eigenstate is occupied.
        Note:
            The sign of the alpha parameter and mass_bandedge are important. If these are negative (as would be expected for the valence band), then they must be passed as negative values to the function.
           
        """
        assert temp > 0, "temperature must be more than 0K"
        if self.band_type == "conduction_band":
            return (1 / ((np.exp((self.energies[0] + self._kane_dispersion(
                k, alpha, mass_bandedge) - fermi_level) / (
                    (boltzmann * temp) / q))) + 1))
        elif self.band_type == "valence_band":
            return 1 - (1 / ((np.exp((self.energies[0] + self._kane_dispersion(
                k, alpha, mass_bandedge) - fermi_level) / (
                    (boltzmann * temp) / q))) + 1))
        elif self.band_type == "unknown":
            raise ValueError("Unable to calculate fermi function as there is partial occupancy of the bands and the band type is unknown. Please set the Segment.band_type attribute manually (options are \"valence_band\" or \"conduction_band\").")
        else:
            raise ValueError("Please set the Segment.band_type attribute (options are \"valence_band\" or \"conduction_band\")")

    def _kane_dispersion(self, k, alpha, mass_bandedge):
        """Helper function for :meth:`~effmass.analysis.Segment.fd`.

        Analytic form of the kane dispersion.
        """
        # returns in eV

        return (_solve_quadratic(
            alpha, 1, [(-k * k) / (2 * mass_bandedge)])[0]) / ev_to_hartree

    def _second_derivative_kane_dispersion(self, k, alpha, mass_bandedge):
        """Helper function for
        :meth:`~effmass.analysis.Segment.mass_integration`.

        Analytic form of the second derivative of the kane dispersion.
        """
        # returns in eV
        return 1 / (mass_bandedge * ((1 + (
            (2 * k * k * alpha) / (mass_bandedge)))**(3 / 2)))

    def weight_integration(self,
                           fermi_level=None,
                           temp=300,
                           alpha=None,
                           mass_bandedge=None,
                           upper_limit=None):
        """Integrates the product of the fermi-dirac distribution
        along the one-dimensional slice of k-space defined by
        :class:`~effmass.analysis.Segment` (up to
        :meth:`~effmass.analysis.Segment.explosion_index`).

        Args:
            fermi_level (float, optional ): Fermi level (eV) to be used in Fermi-dirac statistics. Defaults to :attr:`~effmass.analysis.Segment.fermi_energy`.
            temp (float, optional): The temperature (K) to be used in Fermi-Dirac statistics. Defaults to 300.
            upper_limit (float, optional): The integration upper limit (bohr$^{-1}$). Defaults to where the Kane quasi-linear dispersion is no longer valide, defined by :meth:`~effmass.analysis.Segment.explosion_index`.

        Returns:
            float: A normalisation factor for :meth:`~effmass.analysis.Segment.mass_integration`.
        Note:
            The sign of the alpha parameter and mass_bandedge are important. If these are negative (as would be expected for the valence band), then they must be passed as negative values to the function.
       
        """
        alpha = self.alpha() if alpha is None else alpha
        mass_bandedge = self.kane_mass_band_edge(
        ) if mass_bandedge is None else mass_bandedge
        fermi_level = self.fermi_energy if fermi_level is None else fermi_level
        upper_limit = self.dk_bohr[
            self.explosion_index()] if upper_limit is None else upper_limit
        result = integrate.quad(
            self._weight_integrand,
            0,
            upper_limit,
            args=(fermi_level, temp, alpha, mass_bandedge))
        return result

    def _weight_integrand(self, k, fermi_level, temp, alpha, mass_bandedge):
        """Helper function for
        :meth:`~effmass.analysis.Segment.weight_integration`.

        A function for the weight integrand.
        """
        return self.fd(k, fermi_level, temp, alpha, mass_bandedge)

    def optical_effmass_kane_dispersion(self,
                                        fermi_level=None,
                                        temp=300,
                                        alpha=None,
                                        mass_bandedge=None,
                                        upper_limit=None):
        r"""
        Calculates the optical effective mass, with the dispersion approximated by a Kane quasi-linear function.

        This optical effective mass is defined as:
            ..math::
                \frac{1}{m_o} = \frac{\int f(E_k(k),T) \frac{\delta^2 E_k(k)}{\delta k^2} dk}{\int f(E_k(k),T)  dk}

        where the integral is along the one-dimensional slice of k-space defined by :class:`~effmass.analysis.Segment` (up to :meth:`~effmass.analysis.Segment.explosion_index`) and :math:`f(E_k,T)` is the Fermi-Dirac distribution. E_k(k) is the Kane dispersion:
            ..math::
                \frac{\hbar ^2}{2m_{tb}} = E(1+\alpha E)

        where the transport mass at bandedge (:math:`m_{tb}`) is calculated using :meth:`effmass.analysis.Segment.kane_mass_band_edge` and the alpha parameter is calculated using :meth:`effmass.analysis.Segment.alpha`.
        
        Args:
            fermi_level (float, optional ): Fermi level (eV) to be used in Fermi-dirac statistics. Defaults to :attr:`~effmass.analysis.Segment.fermi_energy`.
            temp (float, optional): The temperature (K) to be used in Fermi-Dirac statistics. Defaults to 300.
            alpha (float, optional): The alpha parameter of the Kane dispersion (hartree$^{-1}$)
            mass_bandedge: The mass at bandedge parameter of the Kane dispersion (units electron mass).
            upper_limit (float, optional): The integration upper limit (bohr$^{-1}$). Defaults to where the Kane quasi-linear dispersion is no longer valide, defined by :meth:`~effmass.analysis.Segment.explosion_index`.

        Returns:
            float: The optical effective mass (units of electron mass) of the :class:`~effmass.analysis.Segment`.

        Note:
           The sign of the alpha parameter and mass_bandedge are important. If these are negative (as would be expected for the valence band), then they must be passed as negative values to the function.
        """

        fermi_level = self.fermi_energy if fermi_level is None else fermi_level

        if ((fermi_level) > (self.energies[self.explosion_index()])):
            if (self.band_type == "conduction_band"):
                print(
                    "Fermi level {} is beyond the energy range where the Kane dispersion is valid.".
                    format(fermi_level))

        if (fermi_level) < (self.energies[self.explosion_index()]):
            if self.band_type == "valence_band":
                print(
                    "Fermi level {} is beyond the energy range where the Kane dispersion is valid.".
                    format(fermi_level))

        top = self.mass_integration(
            fermi_level=fermi_level,
            temp=temp,
            alpha=alpha,
            mass_bandedge=mass_bandedge,
            upper_limit=upper_limit)
        bottom = self.weight_integration(
            fermi_level=fermi_level,
            alpha=alpha,
            mass_bandedge=mass_bandedge,
            temp=temp,
            upper_limit=upper_limit)

        assert top[0] != 0, "Unable to calculate the optical effective mass as the Fermi-Dirac distribution at all kpoints is smaller than Python's double precision floats."

        return bottom[0] / top[0]

    ####

    def _polynomial_function(self, polyfit_order=6, polyfit_weighting=True):
        """Helper function which constructs a polynomial function using a least
        squares fit to dispersion data."""
        _check_poly_order(polyfit_order)
        # we know that there is symmetry between k and -k
        negative_dk = [-value for value in self.dk_bohr[::-1]]
        sym_dk = np.concatenate((negative_dk, self.dk_bohr[1:]))
        sym_dE = np.concatenate((self.dE_hartree[::-1], self.dE_hartree[1:]))
        if polyfit_weighting:
            # weight to enable a better fit for the values where it is important
            weighting = self.weighting()
        else:
            weighting = np.ones(len(self.dE_hartree))
        W = np.append(weighting[::-1],
                      weighting[1:])  # as it needs same dimensions as x and y.
        W = np.sqrt(
            np.diag(W))  # to allow dot product between weight and matrix/y.
        # for explanation of the coefficient matrix and least squares fit see https://stackoverflow.com/questions/32260204/numpy-polyfit-passing-through-0
        # for how to incorporate weighting see https://stackoverflow.com/questions/27128688/how-to-use-least-squares-with-weight-matrix-in-python
        # and https://stackoverflow.com/questions/19624997/understanding-scipys-least-square-function-with-irls
        # by eliminating the units matrix of the array we are forcing a zero offset; the fit must pass through 0,0 as is physical
        coeff_matrix = np.vstack([
            sym_dk**i for i in np.arange(polyfit_order + 1)[polyfit_order:0:-1]
        ]).T
        w_coeff_matrix = np.dot(W, coeff_matrix)
        w_sym_dE = np.dot(sym_dE, W)
        coeff = np.append(np.linalg.lstsq(w_coeff_matrix, w_sym_dE)[0],
                          [0])  # remember to set zeroth power to 0!
        function = np.poly1d(coeff)
        # function = np.poly1d(np.polyfit(sym_dk,sym_dE,polyfit_order,w=W)) ----> simple polyfit call for sanity's sake
        return function

    def poly_derivatives(self,
                         polyfit_order=6,
                         polyfit_weighting=True,
                         dk=None):
        """Constructs a polynomial function using a least squares fit to
        Segment dispersion data, then evaluates first and second order
        derivatives.

        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.
            polyfit_weighting (bool, optional): If True, polyfit will be weighted according to occupation of eigenstates. If False, no weighting will be used.
            dk (array, optional): distance (bohr :math:^{-1}) from extrema in reciprocal space at which to evaluate first and second order derivatives. Defaults to 100 points evenly distributed across the whole segment.

        Returns:
            tuple: A tuple containing a 1d array of first derivatives and 1d array of second derivatives, evaluated at dk: ([dedk],[d2edk2])
        """
        function = self._polynomial_function(
            polyfit_order=polyfit_order, polyfit_weighting=polyfit_weighting)
        if dk is None:
            dk = np.linspace(self.dk_bohr[0], self.dk_bohr[-1], 100)
        dedk = function.deriv(m=1)(dk)
        d2edk2 = function.deriv(m=2)(dk)

        return dedk, d2edk2

    def poly_fit(self, polyfit_order=6, polyfit_weighting=True):
        """Constructs a polynomial function using a least squares fit to
        :class:`~effmass.analysis.Segment` dispersion data, then evaluates at
        100 points evenly distributed across
        :attr:`~effmass.analysis.Segment.dk_bohr`.

        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.
            polyfit_weighting (bool, optional): If True, polyfit will be weighted according to occupation of eigenstates. If False, no weighting will be used.

        Returns:
            array: 1d array containing energies (hartree).
        """
        function = self._polynomial_function(
            polyfit_order=polyfit_order, polyfit_weighting=polyfit_weighting)
        values = np.polyval(
            function, np.linspace(self.dk_bohr[0], self.dk_bohr[-1], 100))
        return values

    def optical_poly_effmass(self, polyfit_order=6):
        r"""
        Calculates the optical effective mass with a polynomial approximation to the dispersion

        This optical effective mass is defined as:
            ..math::
                \frac{1}{m_o} = \frac{\sum_i f(E_i) g(k_i) \frac{\delta^2 E}{\delta k^2}|_i}{\sum f(E_i) g(k_i)}

        where the sum is over eigenstates i contained withing the :class:`~effmass.analysis.Segment`. :math:`f_(E_i)` is the probability of occupation (Fermi-Dirac statistics) and :math:`g(k_i)` is the density of states at that k-point.
        
        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.
        
        Returns:
            float: The optical effective mass (units of electron mass) of the :class:`~effmass.analysis.Segment`.
        """
        inertial_mass = self.inertial_effmass(
            polyfit_order=polyfit_order, dk=self.dk_bohr)
        weighting = self.weighting()
        optical_mass = np.power(
            np.average(np.power(inertial_mass, -1), weights=weighting), -1)
        return optical_mass

    def inertial_effmass(self,
                         polyfit_order=6,
                         dk=None,
                         polyfit_weighting=False):
        r"""
        Calculates the inertial (curvature) effective mass (:math:`\frac{1}{\frac{\delta^2E}{\delta k^2}}`), evaluated at :attr:`~effmass.analysis.Segment.dk_bohr`.

        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.
            dk (array, optional): distance (bohr :math:^{-1}) from extrema in reciprocal space at which to evaluate second order derivatives. Defaults to 100 points evenly distributed across the whole segment.
            polyfit_weighting (bool, optional): If True, polyfit will be weighted according to occupation of eigenstates. If False, no weighting will be used.
        
        Returns:
            array(float). 1d array containing the conductivity effective mass (in units of electron rest mass) evaluated at the points specified in dk.
        """
        dedk, d2edk2 = self.poly_derivatives(
            polyfit_order=polyfit_order,
            dk=dk,
            polyfit_weighting=polyfit_weighting)
        conductivity_mass = [(1 / x) for x in d2edk2]
        return conductivity_mass

    def transport_effmass(self,
                          polyfit_order=6,
                          dk=None,
                          polyfit_weighting=False):
        r"""
        Calculates the transport mass (:math:`\frac{k}{\delta E \delta k}`), evaluated at :attr:`~effmass.analysis.Segment.dk_bohr` .

        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.
            dk (array, optional): distance (bohr :math:^{-1}) from extrema in reciprocal space at which to evaluate first order derivatives. Defaults to 100 points evenly distributed across the whole segment.
            polyfit_weighting (bool, optional): If True, polyfit will be weighted according to occupation of eigenstates. If False, no weighting will be used.
           
        Returns:
            array(float). 1d array containing the transport effective mass (in units of electron rest mass) evaluated at the points specified in dk.
        """
        dedk, d2edk2 = self.poly_derivatives(
            polyfit_order=polyfit_order,
            dk=dk,
            polyfit_weighting=polyfit_weighting)
        # dk=0 for first term gives discontinuity
        if dk is None:
            dk = np.linspace(self.dk_bohr[0], self.dk_bohr[-1], 100)
        transport_mass = [(x / y) for x, y in zip(dk[1:], dedk[1:])]
        transport_mass = np.append(np.array([np.nan]), transport_mass)
        return transport_mass

    def alpha(self, polyfit_order=6, truncate=True):
        r"""
        The transport mass (:math:`\frac{k}{\delta E \delta k}`) is calculated as a function of :attr:`~effmass.analysis.Segment.dk_bohr` and fitted to a straight line. The gradient of this line determines the alpha parameter which is used in the kane dispersion.
        
        See :meth:`effmass.analysis.Segment.transport_effmass`.

        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.
            truncate (bool, optional): If True, data only up to :meth:`effmass.analysis.Segment.explosion_index` is used. If False, alpha is calculated using data for the whole segment. Defaults to True.

        Returns:
            float: The alpha parameter (hartree :math:`^{-1}`).
        """
        if truncate:
            self._check_kanefit_points(polyfit_order=polyfit_order)
        transport_mass = self.transport_effmass(
            polyfit_order=polyfit_order,
            dk=self.dk_bohr,
            polyfit_weighting=False)
        if truncate:
            idx = self.explosion_index(polyfit_order=polyfit_order)
            gradient, intercept = np.polyfit(self.dE_hartree[1:idx + 1],
                                             transport_mass[1:idx + 1], 1)
        else:
            gradient, intercept = np.polyfit(self.dE_hartree[1:],
                                             transport_mass[1:], 1)
        alpha = np.divide(gradient, 2 * intercept)
        return alpha

    def kane_mass_band_edge(self, polyfit_order=6, truncate=True):
        r"""
        The transport mass (:math:`\frac{1}{\delta E \delta k}`) is calculated as a function of :attr:`~effmass.analysis.Segment.dk_bohr` and fitted to a straight line. The intercept of this line with the y-axis gives a transport mass at bandedge which is used as a parameter in the kane dispersion.
        
        See :meth:`effmass.analysis.Segment.transport_effmass`.

        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.
            truncate (bool, optional): If True, data only up to :meth:`effmass.analysis.Segment.explosion_index` is used. If False, alpha is calculated using data for the whole segment. Defaults to True.
        
        Returns:
            float: transport mass at bandedge (in units of electron mass).
        """
        if truncate:
            self._check_kanefit_points(polyfit_order=polyfit_order)
        transport_mass = self.transport_effmass(
            polyfit_order=polyfit_order,
            dk=self.dk_bohr,
            polyfit_weighting=False)
        if truncate:
            idx = self.explosion_index(polyfit_order=polyfit_order)
            gradient, intercept = np.polyfit(self.dE_hartree[1:idx + 1],
                                             transport_mass[1:idx + 1], 1)
        else:
            gradient, intercept = np.polyfit(self.dE_hartree[1:],
                                             transport_mass[1:], 1)
        return intercept  # intercept is the band edge transport mass

    def kane_fit(self, polyfit_order=6):
        r"""
        Calculate the Kane quasi-linear dispersion parameters, then evaluates at 100 points evenly distributed across :attr:`~effmass.analysis.Segment.dk_bohr`.

        The Kane quasi-linear dispersion is described by 
            ..math::
                \frac{\hbar ^2 k^2}{2m_{tb}} = E(1+\alpha E)

        where the transport mass at bandedge (:math:`m_{tb}`) and the alpha parameter are calculated by fitting a linear function to the transport mass :math:`m_t` as a function of energy E.
            ..math::
                m_t = m_{tb}(1+2\alpha E)

        The transport mass :math:`m_t` is calculated by approximating the dispersion with a polynomial function and taking the first derivative, see :meth:`~effmass.analysis.Segment.transport_effmass`.

        Args:
            polyfit_order (int, optional): order of polynomial used to approximate the dispersion. Defaults to 6.
 
        Returns:
            array: 1d array containing energies (hartree).
        """
        self._check_kanefit_points(polyfit_order=polyfit_order)
        alpha = self.alpha(polyfit_order=polyfit_order)
        transport_mass_bandedge = self.kane_mass_band_edge(
            polyfit_order=polyfit_order)
        bandedge_energy = [
            ((k**2)) / (2 * transport_mass_bandedge)
            for k in np.linspace(self.dk_bohr[0], self.dk_bohr[-1], 100)
        ]
        roots = _solve_quadratic(alpha, 1,
                                 [(-1) * ele for ele in bandedge_energy])
        return roots

    def finite_difference_effmass(self):
        """The curvature at the band edge is calculated using a second order
        forward finite difference method. This is then inverted to give an
        effective mass.

        Args:
            None

        Returns:
            float: Bandedge effective mass from finite difference (in units of electron mass).
        """
        x = (self.dE_hartree[2] -
             (2 * self.dE_hartree[1])) / (self.dk_bohr[1]**2)
        mass = 1 / x
        return mass

    def finite_difference_fit(self):
        r"""
        Calculates the curvature at the band edge using a finite difference method and then evaluates the corresponding quadratic dispersion along :attr:`~effmass.analysis.Segment.dk_bohr`.

        See :meth:`effmass.analysis.Segment.finite_difference_effmass`.

        Args:
            None

        Returns:
            list(float): list containing energies (hartree). The energies are calculated at 100 points evenly distributed across :attr:`~effmass.analysis.Segment.dk_bohr` using the quadratic approximation. 
        """
        m_bandedge = self.finite_difference_effmass()
        values = [((k**2) / (2 * m_bandedge))
                  for k in np.linspace(self.dk_bohr[0], self.dk_bohr[-1], 100)]
        return values

    def weighted_leastsq_effmass(self):
        """Fits a parabolic dispersion using the weighted least-squares method
        to all points in the segment, plus those from symmetry, E(k)=E(-k).

        Args:
            None

        Returns:
            float: Curvature effective mass (in units of electron mass)

        Notes:
            weighting is given by the Fermi-Dirac distribution.
        """

        # https://stackoverflow.com/questions/19624997/understanding-scipys-least-square-function-with-irls
        # https://stackoverflow.com/questions/27128688/how-to-use-least-squares-with-weight-matrix-in-python

        negative_dk = [-value for value in self.dk_bohr[::-1]]
        sym_dk = np.concatenate((negative_dk, self.dk_bohr[1:]))
        coeff_matrix = np.vstack([sym_dk**2]).T

        weighting = self.weighting()
        W = np.append(weighting[::-1],
                      weighting[1:])  # as it needs same dimensions as x and y.
        W = np.sqrt(
            np.diag(W))  # to allow dot product between weight and matrix/y.

        sym_dE = np.concatenate((self.dE_hartree[::-1], self.dE_hartree[1:]))
        w_sym_dE = np.dot(sym_dE, W)
        w_coeff_matrix = np.dot(W, coeff_matrix)

        coeff = np.linalg.lstsq(w_coeff_matrix, w_sym_dE)[0]
        mass = 1 / (2 * coeff[0])
        return mass

    def weighted_leastsq_fit(self):
        """Calculates the curvature effective mass using a weighted least-
        squares fit and then evaluates the corresponding parabolic dispersion
        along :attr:`~effmass.analysis.Segment.dk_bohr`.

        Args:
            None

        Returns:
            list(float): list containing energies (hartree). The energies are calculated at 100 points evenly distributed across :attr:`~effmass.analysis.Segment.dk_bohr`.
        """

        m_bandedge = self.weighted_leastsq_effmass()
        values = [((k**2) / (2 * m_bandedge))
                  for k in np.linspace(self.dk_bohr[0], self.dk_bohr[-1], 100)]
        return values

    def five_point_leastsq_effmass(self):
        """Fits a parabolic dispersion using the least-squares method to 5
        points (3 DFT-calculated points + 2 from symmetry).

        Args:
            None

        Returns:
            float: Curvature effective mass (in units of electron mass).

        Notes:
            no weighting is used.
        """

        sym_dk = np.array([
            -self.dk_bohr[2], -self.dk_bohr[1], self.dk_bohr[0],
            self.dk_bohr[1], self.dk_bohr[2]
        ])
        sym_dE = np.array([
            self.dE_hartree[2], self.dE_hartree[1], self.dE_hartree[0],
            self.dE_hartree[1], self.dE_hartree[2]
        ])
        coeff_matrix = np.vstack([sym_dk**2]).T
        coeff = np.linalg.lstsq(coeff_matrix, sym_dE)[0]
        mass = 1 / (2 * coeff[0])
        return mass

    def five_point_leastsq_fit(self):
        """Calculates the curvature effective mass using a parabolic least-
        squares fit and then evaluates the corresponding parabolic dispersion
        along :attr:`~effmass.analysis.Segment.dk_bohr`.

        Args:
            None

        Returns:
            list(float): list containing energies (hartree). The energies are calculated at 100 points evenly distributed across :attr:`~effmass.analysis.Segment.dk_bohr`.
        """
        m_bandedge = self.five_point_leastsq_effmass()
        values = [((k**2) / (2 * m_bandedge))
                  for k in np.linspace(self.dk_bohr[0], self.dk_bohr[-1], 100)]
        return values





#! /usr/bin/env python3

"""
A module for finding the band structure extrema and instantiating 
a :class:`Segment` object for each extrema point.

The extrema are found within an energy range given by the :class:`Settings` class.
Each `Segment` object contains data for kpoints within an energy range given by the :class:`Settings` class. 
"""

import warnings
import numpy as np
import numpy.ma as ma
from effmass import analysis


def _check_partial_occupancy(occupancy):
    """Raises warning if there is partial occupancy of bands.

    Args:
        occupancy (array): Array with shape (number_of_bands,number_of_kpoints). Each row contains occupation number of the eigenstates for a particular band. Values range from 0-1 (spin-polarised) or 0-2 (non-spin-polarised). See :attr:`effmass.inputs.Data.occupancy`.

    Returns:
        None.
    """
    if not np.all(np.in1d(occupancy,([0, 1, 2]))):
        warnings.warn(
            "You have partial occupation numbers in Data.occupancy. You should check that the attributes Data.VBM, Data.CBM and Data.fermi_energy are correct, and if not, set them manually."
        )


def _calc_CBM(occupancy, energy):
    """Finds the minimum unoccupied energy eigenstate (conduction band minimum)
    .

    Args:
        occupancy (array): occupancy of eigenstates with shape (number_of_kpoints, number_of_bands)
        energy (array): energy (eV) of eigenstates with shape (number_of_kpoints, number_of_bands)

    Returns:
        float: the minimum unoccupied energy (eV)
    """
    _check_partial_occupancy(occupancy)
    energy_unoccupied = ma.masked_where(occupancy > 0.5, energy)
    return np.amin(energy_unoccupied)


def _calc_VBM(occupancy, energy):
    """Finds the minimum unoccupied energy eigenstate (valence band maximum).

    Args:
        occupancy (array): occupancy of eigenstates with shape (number_of_kpoints, number_of_bands)
        energy (array): energy (eV) of eigenstates with shape (number_of_kpoints, number_of_bands)

    Returns:
        float: the minimum unoccupied energy (eV)
    """
    _check_partial_occupancy(occupancy)
    energy_occupied = ma.masked_where(occupancy < 0.5, energy)
    return np.amax(energy_occupied)


def calculate_direction(a, b):
    """Calculates the direction vector between two points.

    Args:
        a (list): the position vector of point a.
        b (list): the position vector of point b.

    Returns:
        array: The (unnormalised) direction vector between points a and b. The smallest magnitude of an element is 1 (eg: [1,1,2]).
    """
    difference = np.subtract(a, b)
    if np.count_nonzero(difference) < 1:
        print("The two k-points are equal")
        return np.array([0, 0, 0])

    # we need to find the smallest non-zero value within a-b
    a = np.array(a)
    b = np.array(b)
    direction_masked = ma.masked_equal(
        a - b, 0)  # return array with invalid entries where values are equal
    direction_filled = ma.filled(
        direction_masked, 10
        **6)  # fill invalid elements of array with a large number s
    direction_absolute = np.absolute(
        direction_filled)  # return absolute values of each element
    smallest = np.amin(direction_absolute)
    direction = (
        b - a) / smallest  # use the minimum absolute value as a divisor a-b
    if -1 in direction:
        direction = np.multiply(direction, -1)
    return direction


def change_direction(kpoints, kpoint_indices):
    """Finds the index of the kpoint (if any) where there is a change of
    direction in reciprocal space.

    Args:
        kpoints (array): array of kpoints with shape (number_of_kpoints, 3). Each row contains the fractional coordinates of a kpoint [kx,ky,kz]. See :attr:`effmass.inputs.Data.kpoints`.
        kpoint_indices (list (int)): the kpoint indices over which to check for change in direction

    Returns:
        int: the index of the kpoint where there is a change of direction. If there is no change of direction, returns None.
    """
    new_direction = calculate_direction(kpoints[kpoint_indices[0]],
                                        kpoints[kpoint_indices[1]])
    change_index = None
    for i, value in enumerate(kpoint_indices[:-1]):
        old_direction = new_direction
        new_direction = calculate_direction(kpoints[kpoint_indices[i]],
                                            kpoints[kpoint_indices[i + 1]])
        if np.linalg.norm(new_direction - old_direction) > 0.005:
            change_index = i + 1
            break
    return change_index


def find_extrema_indices(Data, Settings):
    """Finds the kpoint index and band index of the minimum/maximum energy
    turning points within :attr:`effmass.inputs.Settings.energy_range` of the
    conduction band minimum (:attr:`effmass.inputs.Data.CBM`) / valence band
    maximum (:attr:`effmass.inputs.Data.VBM`).

    Args:
        Data (Data): instance of the :class:`Data` class.
        Settings (Settings): instance of the :class:`Settings` class.

    Returns:
        array: A 2-dimensional array of size (number of extrema, 2). Each row corresponds to an extrema and contains [:attr:`efmmas.inputs.Data.bands` index, :attr:`effmass.inputs.Data.kpoints` index].
    """

    energy_CB = ma.masked_where(Data.energies < Data.CBM, Data.energies)
    energy_VB = ma.masked_where(Data.energies > Data.VBM, Data.energies)

    # returns array of energies within energy_range
    holes_in_range = ma.masked_where(
        np.absolute(energy_VB - Data.VBM) >
        Settings.extrema_search_depth, energy_VB)
    VB_maxima = ma.masked_where(
        _mark_maxima(holes_in_range) == 0, holes_in_range)

    # returns array of energies within energy_range
    electrons_in_range = ma.masked_where(
        np.absolute(energy_CB - Data.CBM) >
        Settings.extrema_search_depth, energy_CB)
    CB_minima = ma.masked_where(
        _mark_minima(electrons_in_range) == 0, electrons_in_range)

    # Returns a 2D array of band numbers and k-points for extrema in the correct energy range.
    # At a given k-point there may be multiple bands with highest energy
    extrema_position = np.append(
        np.argwhere(VB_maxima.mask == 0),
        np.argwhere(CB_minima.mask == 0),
        axis=0)
    return extrema_position


def _mark_maxima(holes_array):
    """Helper function which takes an array of hole energies and returns an
    array for masking those which are not maxima."""
    # create array to append to
    not_maxima = []
    width = holes_array.shape[1]
    height = holes_array.shape[0]

    # handle edge cases
    for band in range(0, height):
        if (holes_array[band, 1] >= holes_array[band, 0]):
            not_maxima.append([band, 0])

    for band in range(0, height):
        if (holes_array[band, -2] >= holes_array[band, -1]):
            not_maxima.append([band, -1])

    # find the bands where numbers either side aren't bigger (the maxima)
    # will also include inflexion but needed to get HSP's
    for band in range(0, height):
        for k in range(1, width - 1):
            if (holes_array[band, k - 1] >= holes_array[band, k]
                    or holes_array[band, k + 1] >= holes_array[band, k]):
                not_maxima.append([band, k])

    # Need to assign false after inspection
    # or it screws with the algorithm
    for row in not_maxima:
        holes_array[row[0], row[1]] = False

    # returns matrix with all non-maxima marked false
    return holes_array


def _mark_minima(electrons_array):
    """Helper function which takes an array of electron energies and returns an
    array for masking those which are not minima."""
    # create array to append to
    not_minima = []
    width = electrons_array.shape[1]
    height = electrons_array.shape[0]

    # handle edge cases
    for band in range(0, height):
        if (electrons_array[band, 1] <= electrons_array[band, 0]):
            not_minima.append([band, 0])

    for band in range(0, height):
        if (electrons_array[band, -2] <= electrons_array[band, -1]):
            not_minima.append([band, -1])

    # find the bands where numbers either side aren't bigger (the maxima)
    for band in range(0, height):
        for k in range(1, width - 1):
            if (electrons_array[band, k - 1] <= electrons_array[band, k] or
                    electrons_array[band, k + 1] <= electrons_array[band, k]):
                not_minima.append([band, k])

    # Need to assign false after inspection
    # or it screws with the algorithm
    for row in not_minima:
        electrons_array[row[0], row[1]] = False

    # returns matrix with all non-maxima marked false
    return electrons_array


def _within(band_index, kpoint_index, trial_kpoint_index, Settings, Data):
    """Helper function which checks whether trial_kpoint_index is less than
    :attr:`effmass.inputs.Settings.energy_range`.

    away from kpoint_index for a given band
    """
    if (np.absolute(Data.energies[band_index, trial_kpoint_index] -
                    Data.energies[band_index, kpoint_index])
        ) < Settings.energy_range:
        return True
    else:
        return False


def _kpoints_after(band_index, kpoint_index, Settings, Data):
    """Helper function that returns eigenstates which are 1) on the same band
    2) before the given kpoint in the bandstructure route and 3) within
    :attr:`effmass.inputs.Settings.energy_range`."""
    trial_kpoint_index = kpoint_index + 1
    kpoint_after = [kpoint_index]
    while (trial_kpoint_index < Data.number_of_kpoints
           and _within(band_index, kpoint_index, trial_kpoint_index, Settings,
                       Data) is True):
        kpoint_after.append(trial_kpoint_index)
        trial_kpoint_index = trial_kpoint_index + 1

    return kpoint_after


def _kpoints_before(band_index, kpoint_index, Settings, Data):
    """Helper function that returns eigenstates which are 1) on the same band
    2) before the given eigenstate in the route through reciprocal space 3)
    within :attr:`effmass.inputs.Settings.energy_range`."""
    trial_kpoint_index = kpoint_index - 1
    kpoint_before = [kpoint_index]
    while trial_kpoint_index >= 0 and _within(band_index, kpoint_index,
                                              trial_kpoint_index, Settings,
                                              Data) is True:
        kpoint_before.append(trial_kpoint_index)
        trial_kpoint_index = trial_kpoint_index - 1

    return kpoint_before


def get_kpoints_before(band_index,
                       kpoint_index,
                       Settings,
                       Data,
                       truncate_dir_change=True):
    """For a given eigenstate, finds eigenstates which 1) belong to the same
    band 2) come before the given eigenstate in the route through reciprocal
    space 3) are within :attr:`effmass.inputs.Settings.energy_range`.

    Args:
        band_index (int): index of :attr:`effmass.inputs.Data.bands`.
        kpoint_index (int): index of :attr:`effmass.inputs.Data.kpoints`.
        Settings (Settings): instance of the :class:`Settings` class.
        Data (Data): instance of the :class:`Data` class.
        truncate_dir_change (bool): If True, truncates eigenstates when there is a change in direction. If False, there is no truncation. Defaults to True.

    Returns:
        list(int): indices of :attr:`effmass.inputs.Data.kpoints`.
    """
    kpoints_before = _kpoints_before(band_index, kpoint_index, Settings, Data)

    if len(kpoints_before) > 2:
        change_index = change_direction(Data.kpoints, kpoints_before)
        if change_index:
            if truncate_dir_change is True:
                kpoints_before = kpoints_before[:change_index]
                if len(kpoints_before) < 3:
                    kpoints_before = None
    else:
        kpoints_before = None

    return kpoints_before


def get_kpoints_after(band_index,
                      kpoint_index,
                      Settings,
                      Data,
                      truncate_dir_change=True):
    """For a given eigenstate, finds eigenstates which 1) belong to the same
    band 2) come after the given eigenstate in the route through reciprocal
    space 3) are within :attr:`effmass.inputs.Settings.energy_range`.

    Args:
        band_index (int): index of :attr:`effmass.inputs.Data.bands`.
        kpoint_index (int): index of :attr:`effmass.inputs.Data.kpoints`.
        Settings (Settings): instance of the :class:`Settings` class.
        Data (Data): instance of the :class:`Data` class.
        truncate_dir_change (bool): If True, truncates eigenstates when there is a change in direction. If False, there is no truncation. Defaults to True.

    Returns:
        list(int): indices of :attr:`effmass.inputs.Data.kpoints`.
    """
    kpoints_after = _kpoints_after(band_index, kpoint_index, Settings, Data)

    # now make sure that there is not a change in direction for this set of kpoints and that long enough
    if len(kpoints_after) > 2:
        change_index = change_direction(Data.kpoints, kpoints_after)
        if change_index:
            if truncate_dir_change is True:
                kpoints_after = kpoints_after[:change_index]
                if len(kpoints_after) < 3:
                    kpoints_after = None
    else:
        kpoints_after = None

    return kpoints_after


def generate_segments(Settings, Data, bk=None, truncate_dir_change=True):
    """Generates a list of Segment objects.

    Args:
        Settings (Settings): instance of the :class:`Settings` class.
        Data (Data): instance of the :class:`Data` class.
        truncate_dir_change (bool): If True, truncates eigenstates when there is a change in direction. If False, there is no truncation. Defaults to True.
        bk (list(int)): To manually set an extrema point, in format [:attr:`efmmas.inputs.Data.bands` index, :attr:`effmass.inputs.Data.kpoints` index]. Defaults to None.
   
   Returns:
        list(Segments): A list of :class:`Segment` objects.
    """
    if bk:
        extrema_array = bk
    else:
        extrema_array = find_extrema_indices(Data, Settings)
    kpoints_list = []
    band_list = []
    for band, kpoint in extrema_array:
        kpoints_before = get_kpoints_before(
            band,
            kpoint,
            Settings,
            Data,
            truncate_dir_change=truncate_dir_change)
        kpoints_after = get_kpoints_after(
            band,
            kpoint,
            Settings,
            Data,
            truncate_dir_change=truncate_dir_change)
        if kpoints_before:
            kpoints_list.append(kpoints_before)
            band_list.append(band)
        if kpoints_after:
            kpoints_list.append(kpoints_after)
            band_list.append(band)
    segments = [
        analysis.Segment(Data, band, kpoints)
        for band, kpoints in zip(band_list, kpoints_list)
    ]
    return segments



#! /usr/bin/env python3
"""
A module for analysing DOSCAR data.
"""

import numpy as np


def _check_integrated_dos_loaded(Data):
    """Helper function to check if :attr:`~effmass.inputs.Data.integrated_dos`
    is loaded.

    Args:
        Data (Data): Instance of the :class:`Data` class.

    Returns:
        None.
    """
    assert Data.integrated_dos != [], "Data.integrated_dos is empty. Please set attribute, perhaps using Data.parse_DOSCAR, and try again."
    return


def _check_dos_loaded(Data):
    """Helper function to check if :attr:`~effmass.inputs.Data.dos` is loaded.

    Args:
        Data (Data): Instance of the :class:`Data` class.

    Returns:
        None.
    """
    assert Data.integrated_dos != [], "Data.dos is empty. Please set attribute, perhaps using Data.parse_DOSCAR, and try again."
    return


def find_dos_VBM_index(Data):
    """Finds the lowest index of the
    :attr:`~effmass.inputs.Data.integrated_dos` array where the energy exceeds
    :attr:`~effmass.inputs.Data.VBM`.

    Args:
        Data (Data): Instance of the Data class.

    Returns:
        int: the lowest index of the :attr:`~effmass.inputs.Data.integrated_dos` array where the energy exceeds :attr:`~effmass.inputs.Data.VBM`.
    """
    _check_integrated_dos_loaded(Data)

    for i in range(len(Data.integrated_dos)):
        if Data.VBM < Data.integrated_dos[i][0]:
            if Data.dos[i][1] == 0.0:
                return i


def find_dos_CBM_index(Data):
    """Finds the highest index of the
    :attr:`~effmass.inputs.Data.integrated_dos` array where the energy is less
    than :attr:`~effmass.inputs.Data.CBM`.

    Args:
        Data (Data): Instance of the Data class.

    Returns:
        int: the highest index of the :attr:`~effmass.inputs.Data.integrated_dos` array where the energy is less than :attr:`~effmass.inputs.Data.CBM`.
    """
    _check_integrated_dos_loaded(Data)

    for i in range(len(Data.integrated_dos))[::-1]:
        if Data.CBM > Data.integrated_dos[i][0]:
            if Data.dos[i][1] == 0.0:
                return i


def electron_fill_level(Data, volume, concentration, CBM_index):
    r"""
    Finds the energy to which a given electron concentration will fill the density of states in :attr:`~effmass.inputs.Data.integrated_dos`.
    
    Uses linear interpolation to estimate the energy between two points given in the DOSCAR.

    Args:
        Data (Data): Instance of the :class:`Data` class. 
        volume (float): volume of the unit cell in angstrom :math:`^3`.
        concentration (float): electron concentration in cm :math:`^{-3}`.
        CBM_index (int): highest index of the :attr:`~effmass.inputs.Data.integrated_dos` array where the energy is less than :attr:`~effmass.inputs.Data.CBM`.

    Returns:
        float: the energy (eV, referenced from the CBM) to which the electrons will fill. For the case where the concentration specified would fill all states specified by :attr:`~effmass.inputs.Data.integrated_dos`, None is returned.

    Notes:
        The precision of the result will depend upon the energy resolution in the DOSCAR.
    """
    _check_integrated_dos_loaded(Data)
    CBM_index = CBM_index
    states_per_unit_cell = volume * 1E-30 * concentration * 1E6
    assert (
        states_per_unit_cell < np.absolute(Data.integrated_dos[-1][1] -
                                           Data.integrated_dos[CBM_index][1])
    ), "the concentration specified would fill all available energy states"

    upper_index = len(Data.integrated_dos) - 1
    lower_index = CBM_index

    # this function is made a little more complicated because the dos can be a step function
    # therefore, we cannot interpolate between consecutive indices, but need to find the range of the step.

    for i in range(CBM_index + 1, len(Data.integrated_dos)):
        if states_per_unit_cell < np.absolute(
                Data.integrated_dos[i][1] - Data.integrated_dos[CBM_index][1]):
            upper_index = i  # marks the maximum energy for this concentration
            break

    for i in range(1, upper_index - CBM_index):
        if Data.integrated_dos[upper_index -
                               1][1] - Data.integrated_dos[upper_index - 1 -
                                                           i][1] != 0:
            lower_index = upper_index - i  # marks the minimum energy for this concentration
            break

    # linear interpolation
    proportion = (states_per_unit_cell - (Data.integrated_dos[lower_index][1] -
                                          Data.integrated_dos[CBM_index][1])
                  ) / np.absolute(Data.integrated_dos[upper_index][1] -
                                  (Data.integrated_dos[lower_index][1]))
    energy = ((Data.integrated_dos[lower_index][0] -
               Data.integrated_dos[CBM_index][0]) +
              (Data.integrated_dos[upper_index][0] -
               Data.integrated_dos[lower_index][0]) * proportion)

    # where lower index has been calculated to be below the CBM, as the concentration is smaller than the first step so set lower energy equal to CBM (0eV)
    if Data.integrated_dos[lower_index][0] - Data.integrated_dos[CBM_index][0] < 0:
        energy = (Data.integrated_dos[upper_index][0] -
                  Data.integrated_dos[CBM_index][0]) * proportion

    return energy


def hole_fill_level(Data, volume, concentration, VBM_index):
    r"""
    Finds the energy to which a given hole concentration will fill the density of states in :attr:`~effmass.inputs.Data.integrated_dos`.
    
    Uses linear interpolation to estimate the energy between two points given in the DOSCAR.

    Args:
        Data (Data): Instance of the :class:`Data` class. 
        volume (float): volume of the unit cell in angstrom :math:`^3`.
        concentration: hole concentration in cm :math:`^{-3}`.
        VBM_index (int): lowest index of the :attr:`~effmass.inputs.Data.integrated_dos` array where the energy is more than than :attr:`~effmass.inputs.Data.VBM`.

    Returns:
        float: the energy (eV, referenced from the VBM) to which the holes will fill. For the case where the concentration specified would fill all states specified by :attr:`~effmass.inputs.Data.integrated_dos`, None is returned.

    Notes:
        The precision of the result will depend upon the energy resolution in the DOSCAR. 
    """
    _check_integrated_dos_loaded(Data)
    VBM_index = VBM_index
    states_per_unit_cell = volume * 1E-30 * concentration * 1E6
    assert (
        states_per_unit_cell < np.absolute(Data.integrated_dos[0][1] -
                                           Data.integrated_dos[VBM_index][1])
    ), "the concentration specified would fill all available energy states"

    upper_index = 0
    lower_index = VBM_index

    # this function is made a little more complicated because the dos can be a step function
    # therefore, we cannot interpolate between consecutive indices, but need to find the range of the step.

    for i in range(0, VBM_index)[::-1]:
        if states_per_unit_cell < np.absolute(
                Data.integrated_dos[i][1] -
                Data.integrated_dos[VBM_index - 1][1]):
            upper_index = i  # marks the maximum hole energy for this concentration
            break

    for i in range(1, VBM_index):
        if Data.integrated_dos[upper_index +
                               1][1] - Data.integrated_dos[upper_index + 1 +
                                                           i][1] != 0:
            lower_index = upper_index + i  # marks the minimum hole energy for this concentration
            break

    # linear interpolation
    proportion = (states_per_unit_cell -
                  np.absolute(Data.integrated_dos[lower_index][1] -
                              Data.integrated_dos[VBM_index][1])
                  ) / np.absolute(Data.integrated_dos[upper_index][1] -
                                  (Data.integrated_dos[lower_index][1]))
    energy = ((Data.integrated_dos[lower_index][0] -
               Data.integrated_dos[VBM_index][0]) -
              (Data.integrated_dos[lower_index][0] -
               Data.integrated_dos[upper_index][0]) * proportion)

    # where lower index has been calculated to be above the CBM, as the concentration is smaller than the first step and so set lower energy equal to VBM (0eV)
    if Data.integrated_dos[lower_index][0] - Data.integrated_dos[VBM_index][0] > 0:
        energy = (Data.integrated_dos[upper_index][0] -
                  Data.integrated_dos[VBM_index][0]) * proportion

    return energy



#! /usr/bin/env python3

"""
A module for plotting and summarising segments information, 
density-of-states information and effective mass analysis.

"""
import matplotlib
matplotlib.use('Agg',warn=False)

import matplotlib.pyplot as plt
import numpy as np
from adjustText import adjust_text

from effmass import ev_to_hartree
from effmass.dos import _check_integrated_dos_loaded
from effmass.dos import _check_dos_loaded
from effmass.analysis import _check_poly_order


def plot_segments(Data, Settings, segments):
    """Plots bandstructure overlaid with the DFT-calculated points for each Segment
    instance. Each Segment is labelled with it's direction in reciprocal space
    and index number from the segments argument.

    Args:
        Data (Data): instance of the :class:`Data` class.
        Settings (Settings): instance of the :class:`Settings` class.
        segments (list(Segment)): A list of instances of the :class:`Segment` class.

    Returns:
        Figure, Axes: tuple containing instance of the `matplotlib.pyplot.figure <https://matplotlib.org/api/figure_api.html>`_ class and `matplotlib.pyplot.axes <https://matplotlib.org/api/axes_api.html>`_ class.

    Notes:
        The x-axis of the plot is not to scale.
    """
   
    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111)

    [ax.plot(range(len(Data.energies[i])), Data.energies[i] - Data.VBM) for i in range(len(Data.energies))]
    points = [ax.scatter(segments[i].kpoint_indices, segments[i].energies - Data.VBM) for i in range(len(segments))]
    texts = [ax.text(segments[i].kpoint_indices[-1], segments[i].energies[-1] - Data.VBM, str(i)+", "+str(np.round(segments[i].direction,3))) for i in range(len(segments))]    
    adjust_text(texts, arrowprops=dict(arrowstyle='->', color='red'), autoalign='x', force_text=(0.01,0.025))

    ax.set_ylim([
        -(Settings.extrema_search_depth + Settings.energy_range + 1),
        (Data.CBM - Data.VBM) +
        (Settings.extrema_search_depth + Settings.energy_range + 1)
    ])

    return fig, ax
    

def plot_integrated_dos(Data):
    """Plots integrated density of states (states/unit-cell) against energy
    (eV).

    Args:
        Data (Data): instance of the :class:`Data` class.

    Returns:
        Figure, Axes: tuple containing instance of the `matplotlib.pyplot.figure <https://matplotlib.org/api/figure_api.html>`_ class and `matplotlib.pyplot.axes <https://matplotlib.org/api/axes_api.html>`_ class.

    Notes:
        The valence band maximum is set to 0 eV.
    """
    _check_integrated_dos_loaded(Data)

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111)
    energy = [x[0] - Data.VBM for x in Data.integrated_dos]
    dos_data = [x[1] for x in Data.integrated_dos]
    ax.plot(energy, dos_data)
    ax.set_xlabel("Energy, eV")
    ax.set_ylabel("Integrated DOS, states / unit cell")
    ax.axvline(0, linestyle="--")
    ax.axvline(Data.CBM - Data.VBM, linestyle="--")

    return fig, ax


def plot_dos(Data):
    """Plots density of states (states/unit-cell) against energy (eV).

    Args:
        Data (Data): instance of the :class:`Data` class.

    Returns:
        Figure, Axes: tuple containing instance of the `matplotlib.pyplot.figure <https://matplotlib.org/api/figure_api.html>`_ class and `matplotlib.pyplot.axes <https://matplotlib.org/api/axes_api.html>`_ class.

    Notes:
        The valence band maximum is set to 0 eV.
    """
    _check_dos_loaded(Data)

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111)
    energy = [x[0] - Data.VBM for x in Data.dos]
    dos = [x[1] for x in Data.dos]
    ax.plot(energy, dos)
    ax.set_xlabel("Energy, eV")
    ax.set_ylabel("DOS")
    ax.axvline(0, linestyle="--")
    ax.axvline(Data.CBM - Data.VBM, linestyle="--")

    return fig, ax


def print_results(segment, data, settings, polyfit_order=None):
    
    polyfit_order = settings.degree_bandfit if polyfit_order is None else polyfit_order
    _check_poly_order(polyfit_order)


    print(segment.band_type, segment.direction)
    print("3-point finite difference mass is {:.2f}".format(
        segment.finite_difference_effmass()))
    print("3-point parabolic mass is {:.2f}".format(
        segment.five_point_leastsq_effmass()))
    try: 
        print("weighted parabolic mass is {:.2f}".format(
            segment.weighted_leastsq_effmass()))
    except AssertionError as e:
        print ("----------\n")
        print (e)
        print ("\n-----------")

    try:
        print("alpha is {:.2f} 1/eV".format(
            segment.alpha(polyfit_order=polyfit_order) * ev_to_hartree))
        print("kane mass at bandedge is {:.2f}".format(
            segment.kane_mass_band_edge(polyfit_order=polyfit_order)))
        if segment.explosion_index(polyfit_order=polyfit_order) == len(
                segment.dE_eV):
            print(
                "the Kane quasi linear approximation is valid for the whole segment"
            )
        else:
            print("the Kane quasi-linear approximation is valid until {:.2f} eV".
                  format(segment.dE_eV[segment.explosion_index(
                      polyfit_order=polyfit_order)]))
    except AssertionError as e:
        print ("----------\n")
        print (e)
        print ("\n-----------")

    try:
        print("optical mass at band edge (assuming the Kane dispersion) is {:.2f}".
              format(segment.optical_effmass_kane_dispersion()))
    except AssertionError:
        pass

    plt.figure(figsize=(8, 8))
    plt.plot(
        np.linspace(segment.dk_angs[0], segment.dk_angs[-1], 100),
        np.divide(
            segment.poly_fit(
                polyfit_order=polyfit_order, polyfit_weighting=False),
            ev_to_hartree),
        marker="x",
        ms=5,
        label="polynomial order {}".format(polyfit_order))
    plt.plot(
        np.linspace(segment.dk_angs[0], segment.dk_angs[-1], 100),
        np.divide(segment.finite_difference_fit(), ev_to_hartree),
        marker="<",
        ms=5,
        label="finite diff parabolic")
    plt.plot(
        np.linspace(segment.dk_angs[0], segment.dk_angs[-1], 100),
        np.divide(segment.five_point_leastsq_fit(), ev_to_hartree),
        marker="p",
        ms=5,
        label="five point parabolic")

    try:
        plt.plot(
            np.linspace(segment.dk_angs[0], segment.dk_angs[-1], 100),
            np.divide(segment.weighted_leastsq_fit(), ev_to_hartree),
            marker=">",
            ms=5,
            label="weighted parabolic")
    except AssertionError:
        pass

    try:
        plt.plot(
            np.linspace(segment.dk_angs[0], segment.dk_angs[-1], 100),
            np.divide(
                segment.kane_fit(polyfit_order=polyfit_order), ev_to_hartree),
            marker="o",
            ms=5,
            label="Kane quasi-linear")
    except AssertionError as e:
        pass

    plt.xlabel(r"k ($ \AA^{-1} $)")
    plt.ylabel("energy (eV)")
    plt.scatter(segment.dk_angs, segment.dE_eV, marker="x", s=200, label="DFT")
    plt.legend()
    plt.show()

    fig, axes = plot_segments(data, settings, [segment])
    plt.show()

    plt.figure(figsize=(8, 8))
    idx = segment.explosion_index(polyfit_order=polyfit_order)
    plt.scatter(
        segment.dE_hartree[1:idx + 1],
        segment.transport_effmass(
            polyfit_order=polyfit_order,
            dk=segment.dk_bohr,
            polyfit_weighting=False)[1:idx + 1])
    plt.plot([0, segment.dE_hartree[idx]],
             np.polyval(
                 np.polyfit(
                     segment.dE_hartree[1:idx + 1],
                     segment.transport_effmass(
                         polyfit_order=polyfit_order,
                         dk=segment.dk_bohr,
                         polyfit_weighting=False)[1:idx + 1], 1),
                 [0, segment.dE_hartree[idx]]))
    plt.ylabel("transport mass")
    plt.xlabel("energy (hartree)")
    plt.xlim([0, segment.dE_hartree[idx]])
    plt.show()



import os
from effmass import __version__

try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

try:
	import pypandoc
	long_description = pypandoc.convert('README.md', 'rst')
except ImportError:
	long_description = open('README.md').read()

config = {
    'description': 'An effective mass package',
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',
    'author': 'Lucy Whalley',
    'author_email': 'lucywhalley@gmail.com',
    'url': 'https://github.com/lucydot/effmass',
    'download_url': "https://github.com/lucydot/effmass/archive/%s.tar.gz" % (__version__),
    'version': __version__,
    'install_requires': [ 'vasppy>=0.5.0.0', 
			  'scipy',
			  'numpy', 
                          'matplotlib',
                          'adjustText',
                          'codeclimate-test-reporter',
                          'pytest-lazy-fixture',
                          'coverage==4.3.4' ],
    'python_requires': '>=3.6',
    'license': 'MIT',
    'packages': [ 'effmass' ],
    'scripts': [],
    'name': 'effmass'
}

setup(**config)



#! /usr/bin/env python3

"""
A module for storing electronic structure data and user settings.

The module contains a :class:`Data` class which parses OUTCAR and PROCAR files using the `vasppy <https://github.com/bjmorgan/vasppy>`_ package. 
A function for parsing DOSCAR files is also provided. 
A :class:`Settings` class stores analysis parameters set by the user.

"""
from vasppy import procar, outcar
from effmass import extrema
import math
import warnings
import numpy as np


class Settings:
    """Class for setting analysis parameters.

    Attributes:     energy_range (float): energy in kT over which the
    segment extends.     extrema_search_depth (float): energy in kT from
    bandedge over which to search for extrema.     degree_bandfit (int):
    the degree of the polynomial which is used to fit to dispersion data
    when calculating the transport mass.
    """

    def __init__(self,
                 energy_range=0.25,
                 extrema_search_depth=0.025,
                 bandfit=6):
        """Initialises an instance of the Settings class and checks input using
        :meth:`check_settings()`.

        Args:
            energy_range (float): energy in eV over which the segment extends. Defaults to 0.25 eV.
            extrema_search_depth (float): energy in eV from bandedge over which to search for extrema. Defaults to 0.025 eV.
            degree_bandfit (int): the degree of the polynomial which is used to fit to dispersion data when calculating the transport mass.

        Returns:
            None.
        """
        self.energy_range = energy_range
        self.extrema_search_depth = extrema_search_depth
        self.degree_bandfit = bandfit
        self.check_settings()

    def check_settings(self):
        """Check that Settings class attributes are sane.

        Args:
            None.

        Returns:
            None.
        """
        assert (self.energy_range >
                0), "The energy range must be a positive number"
        assert (self.extrema_search_depth >
                0), "The energy depth must be a positive number"
        assert (
            type(self.degree_bandfit) == int and self.degree_bandfit > 1
        ), "The bandfit degree must be a positive integer greater than 1"


class Data():
    r"""
    Class for parsing and storing data from a vasp calculation.

    Attributes:
        spin_channels (int): 1 (non-spin-polarised), 2 (spin-polarised), 4 (spin-orbit coupling).
        number_of_kpoints (int): the number of k-points per band.
        number_of_bands (int): the number of bands.
        number_of_ions (int): the number of ions.
        kpoints (array(float)): 2-dimensional array with shape (number_of_kpoints, 3). Each row contains the fractional coordinates of a kpoint [kx,ky,kz].
        energies (array(float)): 2-dimensional array with shape (number_of_bands,number_of_kpoints). Each row contains energies of eigenstates in eV for a particular band. 
        occupancy (array(float)): 2-dimensional array with shape (number_of_bands,number_of_kpoints). Each row contains occupation number of the eigenstates for a particular band. Values range from 0-1 (spin-polarised) or 0-2 (non-spin-polarised).     
        reciprocal_lattice (list(float)): the reciprocal lattice vectors in format [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]], units Angstrom :math:`^{-1}`.
        CBM (float): the conduction band minimum energy in eV.
        VBM (float): the valence band maximum in eV.
        fermi_energy (float): the fermi energy in eV. Automatically set to the mean of Data.CBM and Data.VBM. 
        dos (array): 2-dimensional array. Each row contains density of states data (units "number of states / unit cell")  at a given energy: [energy(float),dos(float)]. 
        integrated_dos: 2-dimensional array. Each row contains integrated density of states data at a given energy: [energy(float),integrated_dos(float)].
    """

    def __init__(self, outcar_path, procar_path, ignore=0, **kwargs):
        r"""
        Initialises an instance of the :class:`~effmass.inputs.Data` class and checks data using :meth:`check_data`.

        Args:
            outcar_path (str): The path to the OUTCAR file
            procar_path (:obj:`str` or :obj:`list`): The path(s) to one or more PROCAR files.
            
            ignore (int): The number of kpoints to ignore at the beginning of the bandstructure slice through kspace (useful for hybrid calculations where zero weightings are appended to a previous self-consistent calculation).
            **kwargs: Additional keyword arguments for reading the PROCAR file(s). 
 
        Returns: 
            None.
        """
        assert (type(outcar_path) == str), "The OUTCAR path must be a string"
        assert (type(ignore) == int and ignore >= 0
                ), "The number of kpoints to ignore must be a positive integer"

        reciprocal_lattice = outcar.reciprocal_lattice_from_outcar(outcar_path)
        if isinstance(procar_path, list):
            vasp_data = procar.Procar.from_files(procar_path, **kwargs)
        elif isinstance(procar_path, str):
            vasp_data = procar.Procar.from_file(procar_path, **kwargs)
        else:
            raise TypeError('procar_path must be a string or list of strings')

        self.spin_channels = vasp_data.spin_channels
        self.number_of_bands = vasp_data.number_of_bands
        self.number_of_ions = vasp_data.number_of_ions

        number_of_kpoints = vasp_data.number_of_k_points
        vasp_data_energies = np.array( [ band.energy for band in np.ravel( vasp_data.bands ) ] )
        vasp_data_occupancies = np.array( [ band.occupancy for band in np.ravel( vasp_data.bands ) ] )
        if vasp_data.calculation['spin_polarised']: # to account for the change in PROCAR format for calculations with 2 spin channels (1 k-point block ---> 2 k-point blocks)
            energies = np.zeros([self.number_of_bands*2,number_of_kpoints]) # This is a very ugly way to slice 'n' dice. Should avoid creating new array and use array methods instead. But it does the job so will keep for now.
            for i in range(self.number_of_bands):
                energies[i] = vasp_data_energies.reshape(
                                            number_of_kpoints*2,  # factor of 2 for each kpoint block
                                            self.number_of_bands).T[i][:number_of_kpoints]
                energies[self.number_of_bands+i] = vasp_data_energies.reshape(
                                            number_of_kpoints*2,
                                            self.number_of_bands).T[i][number_of_kpoints:]
            occupancy = np.zeros([self.number_of_bands*2,number_of_kpoints])
            for i in range(self.number_of_bands):
                occupancy[i] = vasp_data_occupancies.reshape(
                                                 number_of_kpoints*2,
                                                 self.number_of_bands).T[i][:number_of_kpoints]
                occupancy[self.number_of_bands+i] = vasp_data_occupancies.reshape(
                                                 number_of_kpoints*2,
                                                 self.number_of_bands).T[i][number_of_kpoints:]
        else:
            energies = vasp_data_energies.reshape(
                                            number_of_kpoints,
                                            self.number_of_bands).T
            occupancy = vasp_data_occupancies.reshape(
                                                 number_of_kpoints,
                                                 self.number_of_bands).T
        
        # remove values which are from the self-consistent calculation prior to the bandstructure calculation (workflow for hybrid functionals)
        self.energies = np.delete(energies,list(range(ignore)),1)
        self.occupancy = np.delete(occupancy,list(range(ignore)),1)
        self.number_of_kpoints = number_of_kpoints - ignore
        
        # handle negative occupancy values
        if np.any(self.occupancy < 0):
            warnings.warn("One or more occupancies in your PROCAR file are negative. All negative occupancies will be set to zero.")
            self.occupancy[ self.occupancy < 0 ] = 0.0

        self.kpoints = np.array( [ kp.frac_coords 
            for kp in vasp_data.k_points[ignore:vasp_data.number_of_k_points] ] )
        self.reciprocal_lattice = reciprocal_lattice * 2 * math.pi
        self.CBM = extrema._calc_CBM(self.occupancy, self.energies)
        self.VBM = extrema._calc_VBM(self.occupancy, self.energies)
        self.fermi_energy = (self.CBM + self.VBM) / 2
        self.dos = []
        self.integrated_dos = []
        self.check_data()

    def check_data(self):
        """Check that data class attributes are sane.

        Args:
            None.

        Returns:
            None.

        Notes:
            There is a `sanity_check` method which runs automatically when reading data in using the `vasppy.procar <http://vasppy.readthedocs.io/en/latest/vasppy.html#module-vasppy.procar>`_ module.
        """
        assert (
            ((self.spin_channels == 1) | (self.spin_channels == 2) |
             (self.spin_channels == 4)) is True
        ), "Spin channels must have value 1 (non spin-polarised) or 2 (spin-polarised)"
        assert (type(self.number_of_kpoints) == int
                and self.number_of_kpoints > 0
                ), "The number of kpoints is not a positive integer"
        assert (type(self.number_of_bands) == int and self.number_of_bands > 0
                ), "The number of bands is not a positive integer"
        assert (type(self.number_of_ions) == int and self.number_of_ions > 0
                ), "The number of ions is not a positive integer"
        assert (self.CBM >
                self.VBM), "The CBM energy is lower than than the VBM energy"
        if self.fermi_energy < self.VBM:
            warnings.warn("The fermi energy is lower than the VBM")
        if self.fermi_energy > self.CBM:
            warnings.warn("The fermi energy is higher than the CBM")
        if ((self.occupancy == 0) | (self.occupancy == 1) |
            (self.occupancy == 2)).all() is False:
            warnings.warn("You have partial occupancy of bands")

    def parse_DOSCAR(self, filename='./DOSCAR'):
        """Parses the DOS and integrated DOS from a vasp DOSCAR file.

        Args:
            filename (str, optional): The location and filename of the DOSCAR to read in. Defaults to `'./DOSCAR'`.

        Returns:
            None.

        Notes:
            If the DOS has been sampled at more than 10000 points then this function will break at the expression for `num_data_points`.
            In this case, edit your DOSCAR file so that in the header there is a space preceding the number of points.
        """
        with open(filename, 'r') as f:
            lines = f.readlines()

        num_data_points = int(lines[5].split()[2])
        if len(lines[6].split()) == 5:
            self.dos = np.array([[
                float(x.split()[0]),
                float(x.split()[1]) + float(x.split()[2])
            ] for x in lines[6:num_data_points + 6]])
            self.integrated_dos = np.array([[
                float(x.split()[0]),
                float(x.split()[3]) + float(x.split()[4])
            ] for x in lines[6:num_data_points + 6]])
        elif len(lines[6].split()) == 3:
            self.dos = np.array([[float(x.split()[0]),
                                  float(x.split()[1])]
                                 for x in lines[6:num_data_points + 6]])
            self.integrated_dos = np.array(
                [[float(x.split()[0]),
                  float(x.split()[2])] for x in lines[6:num_data_points + 6]])
        else:
            print("problem parsing DOSCAR")
        return
